@@ -64,6 +64,8 @@ static gboolean m_bUseGtkPlusTheme;
 static gboolean m_bDriveCorePointer = FALSE;
 static gboolean m_bExpandWindow;
 static gboolean m_bUseXkbKbd;
+static gboolean m_bUseAuxDicts = FALSE;
+static const gchar *m_AuxDicts = NULL;
 static GokDockType m_eDockType;
 
 /* This GConfClient will be shared by all gconf_client
@@ -225,6 +227,12 @@ gboolean gok_data_read_settings (void)
 	gok_gconf_get_bool (gconf_client, GOK_GCONF_USE_XKB_KBD,
                                  &m_bUseXkbKbd);
 
+	gok_gconf_get_bool (gconf_client, GOK_GCONF_USE_AUX_DICTS, 
+			         &m_bUseAuxDicts);
+
+        gok_gconf_get_string (gconf_client, GOK_GCONF_AUX_DICTS, 
+			         &m_AuxDicts);
+
 	gok_gconf_get_string (gconf_client, GOK_GCONF_DOCK_TYPE,
 			      &a_gchar);
         m_eDockType = gok_data_dock_type_from_string (a_gchar);
@@ -1202,3 +1210,41 @@ gok_data_set_repeat_rate (gint rate)
 {
 	/* TODO: implement in gconf */
 }
+
+
+/* 
+ * gok_data_get_aux_dictionaries:
+ *
+ * Returns a constant string which specifies a semicolon-delimited
+ * list of fully-qualified paths to word lists.
+ */
+const gchar* 
+gok_data_get_aux_dictionaries (void)
+{
+	return m_AuxDicts;
+}
+
+/* 
+ * gok_data_set_aux_dictionaries:
+ *
+ */
+void
+gok_data_set_aux_dictionaries (const gchar *dictionaries)
+{
+	m_AuxDicts = dictionaries;
+	gok_gconf_set_string (gconf_client, GOK_GCONF_AUX_DICTS,
+			      m_AuxDicts);
+	m_bSettingsChanged = TRUE;
+}
+
+gboolean gok_data_get_use_aux_dictionaries (void)
+{	
+	return m_bUseAuxDicts;
+}
+
+void gok_data_set_use_aux_dictionaries (gboolean use_aux)
+{	
+	m_bUseAuxDicts = use_aux;
+	gok_gconf_set_bool (gconf_client, GOK_GCONF_USE_AUX_DICTS, m_bUseAuxDicts);
+	m_bSettingsChanged = TRUE;
+}
