@@ -86,3 +86,6 @@
 #define GOK_GCONF_FEEDBACK_SOUNDNAME       "soundname"
 
 #define GOK_GCONF_WORKAROUND_TEXT          "GConf seems to have problems accessing directories who do not have any siblings who are keys.  This key exists to ensure that the subdirectories in this directory do have a sibling who is a key - this key."
+
+#define GOK_GCONF_USE_AUX_DICTS            "/apps/gok/use_aux_dicts"
+#define GOK_GCONF_AUX_DICTS                "/apps/gok/aux_dictionaries"
