@@ -32,6 +32,7 @@
 #include "gok-modifier.h"
 #include "gok-scanner.h"
 #include "gok-feedback.h"
+#include "gok-word-complete.h"
 #include "gok-data.h"
 #include "gok-branchback-stack.h"
 #include "gok-modifier-keymasks.h"
@@ -171,6 +172,7 @@ gok_key_get_numlock_mask (Display *display)
 	return _numlock_mask;
 }
 
+
 static int
 gok_key_get_xkb_type_index (XkbDescPtr xkb, KeyCode keycode, int group)
 {
@@ -373,12 +375,16 @@ gok_key_from_xkb_key (GokKey *prevKey, GokKeyboard *pKeyboard, Display *display,
   XkbBoundsRec *pBounds;
   int width, height, len;
   gboolean is_modifier = FALSE;
-  unsigned int numlock_mask = gok_key_get_numlock_mask (display);
+  unsigned int numlock_mask;
+  guint modmask = SPI_KEYMASK_SHIFT | SPI_KEYMASK_SHIFTLOCK | SPI_KEYMASK_ALT | 
+	  SPI_KEYMASK_CONTROL | SPI_KEYMASK_MOD1 | SPI_KEYMASK_MOD2 | SPI_KEYMASK_MOD3 | SPI_KEYMASK_MOD4 |
+	  SPI_KEYMASK_MOD5; 
 
   pKey = gok_key_new (prevKey, NULL, pKeyboard);
   numlock_mask = gok_key_get_numlock_mask (display);
+  modmask |= numlock_mask;
   is_modifier = FALSE;
-  pKey->has_text = TRUE; /* TODO: handle non-textual keyboard keys! */
+  pKey->has_text = TRUE; /* TODO: handle non-textual keyboard key caps! */
   pKey->Top = row;
   pBounds = &pGeom->shapes[keyp->shape_ndx].bounds;
   pKey->Bottom = row + MAX (1, (pBounds->y2 - pBounds->y1)/(pBounds->x2 - pBounds->x1));
@@ -415,38 +421,42 @@ gok_key_from_xkb_key (GokKey *prevKey, GokKeyboard *pKeyboard, Display *display,
   else {
           pKey->FontSizeGroup = FONT_SIZE_GROUP_GLYPH;
   }
-  gok_keylabel_new (pKey, label, NULL);
+  /* TODO: figure out how to specify smaller modmasks */
+  gok_keylabel_new (pKey, label, 0, modmask);
   label = gok_key_label_from_keycode (keycode, display, ShiftMask, 0);
-  gok_keylabel_new (pKey, label, "shift");
+  gok_keylabel_new (pKey, label, ShiftMask, modmask);
   label = gok_key_label_from_keycode (keycode, display, LockMask | ShiftMask, 0);
-  gok_keylabel_new (pKey, label, "shift+capslock");
+  gok_keylabel_new (pKey, label, LockMask | ShiftMask, modmask);
   label = gok_key_label_from_keycode (keycode, display, LockMask, 0);
-  gok_keylabel_new (pKey, label, "capslock");
+  gok_keylabel_new (pKey, label, LockMask, modmask);
   label = gok_key_label_from_keycode (keycode, display, Mod1Mask, 0);
-  gok_keylabel_new (pKey, label, "alt");
+  gok_keylabel_new (pKey, label, SPI_KEYMASK_ALT, modmask);
   label = gok_key_label_from_keycode (keycode, display, ControlMask, 0);
-  gok_keylabel_new (pKey, label, "ctrl");
+  gok_keylabel_new (pKey, label, SPI_KEYMASK_CONTROL, modmask);
   label = gok_key_label_from_keycode (keycode, display, numlock_mask, 0);
-  gok_keylabel_new (pKey, label, "numlock");
+  gok_keylabel_new (pKey, label, numlock_mask, modmask);
   label = gok_key_label_from_keycode (keycode, display, numlock_mask | ShiftMask, 0);
-  gok_keylabel_new (pKey, label, "numlock+shift");
+  gok_keylabel_new (pKey, label, SPI_KEYMASK_SHIFT | numlock_mask, 
+		    modmask);
   label = gok_key_label_from_keycode (keycode, display, numlock_mask | LockMask, 0);
-  gok_keylabel_new (pKey, label, "numlock+capslock");
+  gok_keylabel_new (pKey, label, SPI_KEYMASK_SHIFTLOCK | numlock_mask, 
+		    modmask);
   label = gok_key_label_from_keycode (keycode, display, 
 				      numlock_mask | LockMask | ShiftMask, 0);
-  gok_keylabel_new (pKey, label, "numlock+capslock+shift");
+  gok_keylabel_new (pKey, label, SPI_KEYMASK_SHIFTLOCK | SPI_KEYMASK_SHIFT | numlock_mask, 
+		    modmask);
   label = gok_key_label_from_keycode (keycode, display, Mod3Mask, 0);
-  gok_keylabel_new (pKey, label, "mod2");
+  gok_keylabel_new (pKey, label, SPI_KEYMASK_MOD2, modmask);
   label = gok_key_label_from_keycode (keycode, display, Mod3Mask, 0);
-  gok_keylabel_new (pKey, label, "mod3");
+  gok_keylabel_new (pKey, label, SPI_KEYMASK_MOD3, modmask);
   label = gok_key_label_from_keycode (keycode, display, Mod4Mask, 0);
-  gok_keylabel_new (pKey, label, "mod4");
+  gok_keylabel_new (pKey, label, SPI_KEYMASK_MOD4, modmask);
   label = gok_key_label_from_keycode (keycode, display, Mod5Mask, 0);
-  gok_keylabel_new (pKey, label, "mod5");
-  label = gok_key_label_from_keycode (keycode, display, 0, 1);
-  gok_keylabel_new (pKey, label, "group1");
+  gok_keylabel_new (pKey, label, SPI_KEYMASK_MOD5, modmask);
+  /*  label = gok_key_label_from_keycode (keycode, display, 0, 1);
+  gok_keylabel_new (pKey, label, group1_mask, group1_mask);
   label = gok_key_label_from_keycode (keycode, display, ShiftMask, 1);
-  gok_keylabel_new (pKey, label, "shift+group1");
+  gok_keylabel_new (pKey, label, "shift+group1");*/
   snprintf (keycode_name, 8, "%d", keycode);
   pKey->pOutput = gok_output_new (OUTPUT_KEYCODE, keycode_name, SPI_KEY_PRESSRELEASE);
   return pKey;
@@ -638,6 +648,10 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 	GokOutput* pNewOutput;
 	GokOutput* pOutputTemp;
 	gboolean settings_locked;
+	guint modmask =  
+		SPI_KEYMASK_SHIFT | SPI_KEYMASK_SHIFTLOCK | SPI_KEYMASK_ALT | 
+		SPI_KEYMASK_CONTROL | SPI_KEYMASK_MOD1 | SPI_KEYMASK_MOD2 | SPI_KEYMASK_MOD3 | SPI_KEYMASK_MOD4 |
+		SPI_KEYMASK_MOD5; /* ASSUMPTION: this includes numlock mask */
 	
 	g_assert (pKey != NULL);
 	g_assert (pNode != NULL);
@@ -922,7 +936,22 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 		/* key label */
 		if (xmlStrcmp (pNodeKeyChild->name, (const xmlChar *)"label") == 0)
 		{
-			gok_keylabel_new (pKey, (char *)xmlNodeGetContent (pNodeKeyChild), (char *)xmlGetProp (pNodeKeyChild, (const xmlChar *) "modifier"));
+			guint modifier = 0;
+			gchar *modifier_list = (gchar *) xmlGetProp (pNodeKeyChild, (const xmlChar *) "modifier");
+			gchar **tokens;
+			if (modifier_list) 
+			{
+				gint i = 0;
+				tokens = g_strsplit (modifier_list, "+", 16);
+				while (tokens[i]) {
+					modifier |= gok_modifier_mask_for_name (gok_main_display (),
+										tokens[i]);
+					++i;
+				}
+				g_strfreev (tokens);
+			}
+			gok_keylabel_new (pKey, (char *)xmlNodeGetContent (pNodeKeyChild),
+					  modifier, modmask);
 			pKey->has_text = TRUE;
 		}
 
@@ -1026,7 +1055,7 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 *
 * returns: A pointer to the new key label, NULL if it wasn't created.
 **/
-GokKeyLabel* gok_keylabel_new (GokKey* pKey, gchar* pLabelText, gchar* pModifierText)
+GokKeyLabel* gok_keylabel_new (GokKey* pKey, gchar* pLabelText, guint modifiers, guint mod_mask)
 {
 	GokKeyLabel* pNewLabel;
 	GokKeyLabel* pKeyLabel;
@@ -1035,7 +1064,8 @@ GokKeyLabel* gok_keylabel_new (GokKey* pKey, gchar* pLabelText, gchar* pModifier
 	pNewLabel = (GokKeyLabel*) g_malloc(sizeof(GokKeyLabel));
 	
 	pNewLabel->Text = m_TextEmpty;
-	pNewLabel->Modifier = m_TextEmpty;
+	pNewLabel->modifiers = modifiers;
+	pNewLabel->mod_mask = mod_mask;
 	pNewLabel->pLabelNext = NULL;
 
 	if (pLabelText != NULL)
@@ -1044,12 +1074,6 @@ GokKeyLabel* gok_keylabel_new (GokKey* pKey, gchar* pLabelText, gchar* pModifier
 		strcpy (pNewLabel->Text, pLabelText);
 	}
 	
-	if (pModifierText != NULL)
-	{
-		pNewLabel->Modifier = (gchar*)g_malloc (strlen(pModifierText) + 1);
-		strcpy (pNewLabel->Modifier, pModifierText);
-	}
-	
 	if (pKey != NULL)
 	{
 		pKeyLabel = pKey->pLabel;
@@ -1082,10 +1106,6 @@ void gok_keylabel_delete (GokKeyLabel* pKeyLabel)
 		{
 			g_free (pKeyLabel->Text);
 		}
-		if (pKeyLabel->Modifier != m_TextEmpty)
-		{
-			g_free (pKeyLabel->Modifier);
-		}
 		g_free (pKeyLabel);
 	}
 }
@@ -1101,7 +1121,7 @@ void gok_keylabel_delete (GokKeyLabel* pKeyLabel)
 *
 * returns: TRUE if the key was initialized, FALSE if not.
 **/ 
-gboolean gok_key_add_label (GokKey* pKey, gchar* pLabelText, gchar* pModifier)
+gboolean gok_key_add_label (GokKey* pKey, gchar* pLabelText, guint modifiers, guint mod_mask)
 {
 	GokKeyLabel* pKeyLabel;
 		
@@ -1110,7 +1130,7 @@ gboolean gok_key_add_label (GokKey* pKey, gchar* pLabelText, gchar* pModifier)
 	
 	if (pKey->pLabel == NULL)
 	{
-		gok_keylabel_new (pKey, pLabelText, pModifier);
+		gok_keylabel_new (pKey, pLabelText, modifiers, mod_mask);
 	}
 	else
 	{
@@ -1118,7 +1138,7 @@ gboolean gok_key_add_label (GokKey* pKey, gchar* pLabelText, gchar* pModifier)
 		pKeyLabel = pKey->pLabel;
 		while (pKeyLabel != NULL)
 		{
-			if (strcmp (pKeyLabel->Modifier, pModifier) == 0)
+			if ((pKeyLabel->modifiers == modifiers) && (pKeyLabel->mod_mask == mod_mask))
 			{
 				g_free (pKeyLabel->Text);
 				pKeyLabel->Text = (gchar*)g_malloc (strlen(pLabelText) + 1);
@@ -1131,7 +1151,7 @@ gboolean gok_key_add_label (GokKey* pKey, gchar* pLabelText, gchar* pModifier)
 		
 		if (pKeyLabel == NULL)
 		{
-			gok_keylabel_new (pKey, pLabelText, pModifier);
+			gok_keylabel_new (pKey, pLabelText, modifiers, mod_mask);
 		}
 	}
 		
@@ -1150,7 +1170,7 @@ void gok_key_change_label (GokKey* pKey, gchar* LabelText)
 	g_assert (pKey != NULL);
 	g_assert (LabelText != NULL);
 	
-	gok_key_add_label (pKey, LabelText, MODIFIER_NORMAL);
+	gok_key_add_label (pKey, LabelText, 0, 0);
 	
 	/* change the label text */
 	gok_key_set_button_label (pKey, LabelText);
@@ -1372,12 +1392,12 @@ gok_key_create_image_widget (GokKey* pKey)
 gchar* gok_key_get_label (GokKey* pKey)
 {
 	/* FIXME: make this work properly with multi-group keyboards! */
-	/* FIXME: also, it's pretty inefficient at present, slow on XSun...*/
 	gchar* pTextReturned;
 	gchar* pToken;
 	gchar buffer[150];
 	GokKeyLabel* pKeyLabel;
 	gboolean bFoundLabel;
+	guint mods = gok_spy_get_modmask ();
 	
 	g_assert (pKey != NULL);
 	
@@ -1387,7 +1407,7 @@ gchar* gok_key_get_label (GokKey* pKey)
 	pKeyLabel = pKey->pLabel;
 	while (pKeyLabel != NULL)
 	{
-		if (strcmp (pKeyLabel->Modifier, MODIFIER_NORMAL) == 0)
+		if (pKeyLabel->modifiers == 0)
 		{
 			pTextReturned = pKeyLabel->Text;
 			break;
@@ -1409,36 +1429,15 @@ gchar* gok_key_get_label (GokKey* pKey)
 	pKeyLabel = pKey->pLabel;
 	while (pKeyLabel != NULL)
 	{
-		/* does this label apply to more than one modifier? (e.g. ctrl+shift) */
-		if (strchr (pKeyLabel->Modifier, '+') != 0)
+		if ((pKeyLabel->modifiers) == (mods & pKeyLabel->mod_mask)) 
 		{
-			/* check if all these modifiers are on */
-			strcpy (buffer, pKeyLabel->Modifier);
-			pToken = strtok (buffer, "+");
-			while (pToken != NULL)
-			{
-				if (strcmp (pToken, "group1") ||
-				    (gok_key_get_effective_group () != 1)) {
-					if (gok_modifier_get_state (pToken) == MODIFIER_STATE_OFF)
-					{
-						break;
-					}
-				}
-				
-				pToken = strtok (NULL, "+");
-				
-				if (pToken == NULL)
-				{
-					/* all of the modifiers are on, return this label text */
-					return pKeyLabel->Text;
-				}
-			}
+			return pKeyLabel->Text;
 		}
 		else if ((bFoundLabel == FALSE) &&
-				(strcmp (pKeyLabel->Modifier, MODIFIER_NORMAL) != 0))
+				(pKeyLabel->modifiers != 0))
 		{
 			/* the first modifier state that matches will be used */
-			if (gok_modifier_get_state (pKeyLabel->Modifier) != MODIFIER_STATE_OFF)
+			if (pKeyLabel->modifiers & mods & pKeyLabel->mod_mask)
 			{
 				pTextReturned = pKeyLabel->Text;
 				bFoundLabel = TRUE;
@@ -1662,7 +1661,7 @@ gok_key_update_toggle_state (GokKey *pKey)
 	        gboolean is_active = FALSE; 
 		if (pKey->ModifierName != NULL)
 			is_active = 
-			(gok_modifier_get_state (pKey->ModifierName) != MODIFIER_STATE_OFF);
+			(gok_modifier_get_state (pKey->ModifierName) != 0);
 		else 
 			is_active = pKey->ComponentState.active;
 		if (pKey->ComponentState.radio) 
@@ -1724,7 +1723,7 @@ int gok_key_calculate_font_size (GokKey* pKey, gboolean bWidth, gboolean bHeight
 	
 		increment = 500;
 
-		for (sizeFontWidth = 5000; sizeFontWidth < maxFontSize; sizeFontWidth += increment)
+		for (sizeFontWidth = GOK_MIN_FONT_SIZE; sizeFontWidth < maxFontSize; sizeFontWidth += increment)
 		{
 			/* set the key label to the trial size */
 			gok_key_set_font_size (pKey, sizeFontWidth);
@@ -1774,7 +1773,7 @@ int gok_key_calculate_font_size (GokKey* pKey, gboolean bWidth, gboolean bHeight
 		heightCell = gok_data_get_key_height() - 2 *
 		        gok_key_get_default_border_height (pKey) - 1;
 	
-		for (sizeFontHeight = 5000; sizeFontHeight < maxFontSize; sizeFontHeight += increment)
+		for (sizeFontHeight = GOK_MIN_FONT_SIZE; sizeFontHeight < maxFontSize; sizeFontHeight += increment)
 		{
 			/* set the key label to the trial size */
 			gok_key_set_font_size (pKey, sizeFontHeight);
@@ -2214,3 +2213,59 @@ gok_key_isRepeatable(GokKey* pKey)
 	gok_log_leave();
 	return returncode;
 }
+
+
+static GokOutput *last_wordcomplete_output = NULL;
+
+/**
+* gok_key_wordcomplete_output
+* 
+* Gets the output for a word prediction key.
+*
+* @pKey: Pointer to the word completion key that will be output.
+*
+* returns: A pointer to the output.
+**/
+GokOutput* 
+gok_key_wordcomplete_output (GokKey *pKey, GokWordComplete *complete)
+{
+	gchar* LabelText;
+	gchar *word_part = gok_wordcomplete_get_word_part (complete);
+	g_assert (pKey != NULL);
+
+	/* make sure this is a word completion key */
+	if (pKey->Type != KEYTYPE_WORDCOMPLETE)
+	{
+		gok_log_x ("Hey, this is not a word completion key!\n");
+		return NULL;
+	}
+	
+	/* subtract the part word from the output */
+	LabelText = gok_key_get_label (pKey);
+	if (LabelText == NULL)
+	{
+		return NULL;
+	}
+	
+	if (word_part && (g_utf8_strlen (word_part, -1) < g_utf8_strlen (LabelText, -1)))
+	{
+		gchar *cp;
+                /* 
+		 * note that this won't match if the decompositions don't match. Should we
+		 * decompose to canonical first ? 
+		 */ 
+		cp = strstr (LabelText, word_part); 
+		/* sanity check: is our word_part found in our label? */
+		if (cp) {
+			gint offset = strlen (word_part); /* in bytes, not chars */
+			if (last_wordcomplete_output) gok_output_delete_all (last_wordcomplete_output);
+			last_wordcomplete_output = gok_output_new (OUTPUT_KEYSTRING, cp + offset, SPI_KEY_STRING);
+		}
+		else {
+			/* if not then we are doing a 'replace', handle that via EditableText. */
+			g_warning ("Sorry, string replacement is not yet implemented.");
+		}
+	}
+
+	return last_wordcomplete_output;
+}
