@@ -31,6 +31,8 @@ extern "C" {
 typedef enum {
 	GOK_WORDCOMPLETE_CHAR_TYPE_NORMAL,
 	GOK_WORDCOMPLETE_CHAR_TYPE_BACKSPACE,
+	GOK_WORDCOMPLETE_CHAR_TYPE_COMPOSE, /* prefix composition */
+	GOK_WORDCOMPLETE_CHAR_TYPE_COMBINE, /* postfix composition */
 	GOK_WORDCOMPLETE_CHAR_TYPE_DELIMITER
 } GokWordCompletionCharacterCategory;
 
@@ -74,6 +76,8 @@ struct _GokWordCompleteClass
 	gchar*    (* unichar_push) (GokWordComplete *complete, gunichar letter);
 	GokWordCompletionCharacterCategory (* unichar_get_category) (GokWordComplete *complete, gunichar letter);
 	gchar*    (* get_word_part) (GokWordComplete *complete);
+        gchar*    (* get_aux_dictionary) (GokWordComplete *complete);
+	gboolean  (* set_aux_dictionary) (GokWordComplete *complete, const gchar *dict_path);
 };
 
 gboolean gok_wordcomplete_open           (GokWordComplete *complete, gchar *directory);
@@ -91,6 +95,8 @@ gchar**  gok_wordcomplete_process_and_predict (GokWordComplete *complete, const
 					       gint num_predictions);
 
 gchar*   gok_wordcomplete_get_word_part (GokWordComplete *complete);
+gchar*   gok_wordcomplete_get_aux_dictionaries (GokWordComplete *complete);
+gboolean gok_wordcomplete_set_aux_dictionaries (GokWordComplete *complete, const gchar *dictionary_path);
 
 #ifdef __cplusplus
 }
